require 'verbs'
require 'patriarch'


module Patriarch

  # Patriarch::Behaviours is meant to be included into any class we want to enhance with Patriarch handled behaviours
  # Thus it defines the +included+ hook so we enhance the class in which it is included freely and add to 
  # it some class methods defined into Patriarch::Behaviours::ClassMethods
  # We define there the +add_behaviour+ method that does three things : 
  # * it generates an anonymous module containing tool methods defined ad-hoc in function of the arguments supplied
  # * it includes another module that lives into a generated file in lib/ allowing us to override existing methods or add new ones if we wish
  # * it defines instance methods for the class allowing instances to call behaviour actions
  module Behaviours

    # Exception raised when +add_behaviour+ declaration in a class has wrong syntax
    class AddBehaviourSyntaxError < Exception ; end
    class AddBehaviourDependencyError < Exception ; end    

    class << self

      # Helper function to complete the anonymous module we include later in the process in the enhanced model class
      # @param [Module] anon_module refers to the anonymous module to be completed
      # @param [String] behaviour the behaviour for which we complete the module with tools method
      # @param [Array] acted_on_model_list the list of model classes that i can target with the behaviour
      # @param [Hash] options options can serve to build custom alias for tool methods
      def complete_custom_active_module_bipartite(anon_module,behaviour,acted_on_model_list,options={})
        anon_module.class_eval do

          acted_on_model_list.each do |acted_on_model|

            # Compute names based on conventions
            classic_tool_method_name = "#{acted_on_model.to_s.tableize}_i_#{behaviour}"
            raw_tool_method_name = classic_tool_method_name + "_ids"
            redis_key = "patriarch_" + classic_tool_method_name

            # Define methods with the pattern : items_i_like that returns models and items_i_like_ids that return
            # Redis key has the same radical as the method in class by convention (but has a patriarch_ prefix to avoid collisions with other gems)
            define_method(classic_tool_method_name) do |options={}|
              Patriarch::ToolServices::RedisExtractorService.instance.
                get_models_from_ids(self,acted_on_model.to_s.classify.constantize,raw_tool_method_name,options)
            end
            if options[:as]
              alias_for_classic = classic_tool_method_name.sub(behaviour, options[:as].to_s)
              alias_method alias_for_classic, classic_tool_method_name
            end

            define_method(raw_tool_method_name) do |options={}|
              Patriarch::ToolServices::RedisExtractorService.instance.get_ids_from_sorted_set(self,redis_key,options)
            end
            if options[:as]
              alias_for_raw = raw_tool_method_name.sub(behaviour, options[:as].to_s)
              alias_method alias_for_raw, raw_tool_method_name
            end
          end
        end
      end


       # Helper function to complete the anonymous module we include later in the process in the enhanced model class
      # @param [Module] anon_module refers to the anonymous module to be completed
      # @param [String] behaviour the behaviour for which we complete the module with tools method
      # @param [Array] targeted_by_model_list the list of model classes that can target me with the behaviour
      # @param [Hash] options options can serve to build custom alias for tool methods
      def complete_custom_passive_module_bipartite(anon_module,behaviour,targeted_by_model_list,options={})
        anon_module.class_eval do

          targeted_by_model_list.each do |targeted_by_model|

            # Compute names based on conventions
            progressive_present_relation_type = (Verbs::Conjugator.conjugate behaviour.to_sym, :aspect => :progressive).split(/ /).last
            classic_tool_method_name = "#{targeted_by_model.to_s.tableize}_#{progressive_present_relation_type}_me"
            raw_tool_method_name = classic_tool_method_name + "_ids"
            redis_key = "patriarch_" + classic_tool_method_name

            # Define methods with the pattern : items_i_like that returns models and items_i_like_ids that return
            # Redis key has the same radical as the method in class by convention (but has a patriarch_ prefix to avoid collisions with other gems)
            define_method(classic_tool_method_name) do |options={}|
              Patriarch::ToolServices::RedisExtractorService.instance.
                get_models_from_ids(self,targeted_by_model.to_s.classify.constantize,raw_tool_method_name,options)
            end
            if options[:as]
              progressive_present_relation_type_alias = (Verbs::Conjugator.conjugate options[:as], :aspect => :progressive).split(/ /).last
              alias_for_classic = classic_tool_method_name.sub(progressive_present_relation_type,progressive_present_relation_type_alias)
              alias_method alias_for_classic,  classic_tool_method_name
            end

            define_method(raw_tool_method_name) do |options={}|
              Patriarch::ToolServices::RedisExtractorService.instance.get_ids_from_sorted_set(self,redis_key,options)
            end
            if options[:as]
              progressive_present_relation_type_alias = (Verbs::Conjugator.conjugate options[:as], :aspect => :progressive).split(/ /).last
              alias_for_raw = raw_tool_method_name.sub(progressive_present_relation_type,progressive_present_relation_type_alias)
              alias_method alias_for_raw,  raw_tool_method_name
            end

          end

        end
      end

      # Helper function to complete the anonymous module we include later in the process in the enhanced model class
      # @param [Module] anon_module refers to the anonymous module to be completed
      # @param [String] behaviour the behaviour for which we complete the module with tools method
      # @param [Array] acted_on_model_list the list of model classes that i can target with the behaviour
      # @param [Array] via_model_list the list of model classes that the behaviour use to act through
      # @param [Hash] options options can serve to build custom alias for tool methods
      def complete_custom_active_module_tripartite(anon_module,behaviour,acted_on_model_list,via_model_list,options={})
        anon_module.class_eval do

          acted_on_model_list.each do |acted_on_model|
            via_model_list.each do |via_model|

              # Compute names based on conventions
              # fallen_angels_i_praise_via_love_letters
              target_classic_tool_method_name = "#{acted_on_model.to_s.tableize}_i_#{behaviour}_via_#{via_model.to_s.tableize}"
              target_raw_tool_method_name = target_classic_tool_method_name + "_ids"

              # love_letters_i_use_to_praise_fallen_angels
              medium_classic_tool_method_name = "#{via_model.to_s.tableize}_i_use_to_#{behaviour}_#{acted_on_model.to_s.tableize}"
              medium_raw_tool_method_name = medium_classic_tool_method_name + "_ids"

              redis_key = "patriarch_" + target_classic_tool_method_name

              # Define methods with the pattern : items_i_like that returns models and items_i_like_ids that return
              # Redis key has the same radical as the method in class by convention (but has a patriarch_ prefix to avoid collisions with other gems)
              define_method(target_classic_tool_method_name) do |options={}|
                Patriarch::ToolServices::RedisExtractorService.instance.
                  get_models_from_ids(self,acted_on_model.to_s.classify.constantize,target_raw_tool_method_name,options.merge({:tripartite => true}))
              end
              if options[:as]
                alias_for_classic = target_classic_tool_method_name.sub(behaviour, options[:as].to_s)
                alias_method alias_for_classic, target_classic_tool_method_name
              end

              define_method(target_raw_tool_method_name) do |options={}|
                Patriarch::ToolServices::RedisExtractorService.instance.get_ids_from_sorted_set(self,redis_key,options.merge({:tripartite => true, :protagonist_type  => :target}))
              end
              if options[:as]
                alias_for_target_raw = target_raw_tool_method_name.sub(behaviour, options[:as].to_s)
                alias_method alias_for_target_raw, target_raw_tool_method_name
              end

              #
              define_method(medium_classic_tool_method_name) do |options={}|
                Patriarch::ToolServices::RedisExtractorService.instance.
                  get_models_from_ids(self,via_model.to_s.classify.constantize,medium_raw_tool_method_name,options.merge({:tripartite => true}))
              end
              if options[:as]
                alias_for_medium_classic = medium_classic_tool_method_name.sub(behaviour, options[:as].to_s)
                alias_method alias_for_medium_classic, medium_classic_tool_method_name
              end
              #

              define_method(medium_raw_tool_method_name) do |options={}|
                Patriarch::ToolServices::RedisExtractorService.instance.get_ids_from_sorted_set(self,redis_key,options.merge({:tripartite => true, :protagonist_type  => :medium}))
              end
              if options[:as]
                alias_for_medium_raw = medium_raw_tool_method_name.sub(behaviour, options[:as].to_s)
                alias_method alias_for_medium_raw, medium_raw_tool_method_name
              end

            end
          end

        end
      end

      # Helper function to complete the anonymous module we include later in the process in the enhanced model class
      # @param [Module] anon_module refers to the anonymous module to be completed
      # @param [String] behaviour the behaviour for which we complete the module with tools method
      # @param [Array] targeted_by_model_list the list of model classes that can target me with the behaviour
      # @param [Array] via_model_list the list of model classes that the behaviour use to act through
      # @param [Hash] options options can serve to build custom alias for tool methods
      def complete_custom_passive_module_tripartite(anon_module,behaviour,targeted_by_model_list,via_model_list,options={})
        anon_module.class_eval do

          targeted_by_model_list.each do |targeted_by_model|
            via_model_list.each do |via_model|
              # Compute names based on conventions
              progressive_present_relation_type = (Verbs::Conjugator.conjugate behaviour.to_sym, :aspect => :progressive).split(/ /).last

              actor_classic_tool_method_name  = "#{targeted_by_model.to_s.tableize}_#{progressive_present_relation_type}_me_via_#{via_model.to_s.tableize}"
              actor_raw_tool_method_name      = actor_classic_tool_method_name + "_ids"

              #"love_letters_used_by_monsters_to_praise_me"
              medium_classic_tool_method_name = "#{via_model.to_s.tableize}_used_by_#{targeted_by_model.to_s.tableize}_to_#{behaviour}_me"
              medium_raw_tool_method_name     = medium_classic_tool_method_name + "_ids"

              redis_key = "patriarch_" + actor_classic_tool_method_name

              define_method(actor_classic_tool_method_name) do |options={}|
                Patriarch::ToolServices::RedisExtractorService.instance.
                  get_models_from_ids(self,targeted_by_model.to_s.classify.constantize,actor_raw_tool_method_name,options.merge({:tripartite => true}))
              end
              if options[:as]
                progressive_present_relation_type_alias = (Verbs::Conjugator.conjugate options[:as], :aspect => :progressive).split(/ /).last
                alias_for_actor_classic = actor_classic_tool_method_name.sub(progressive_present_relation_type,progressive_present_relation_type_alias)
                alias_method alias_for_actor_classic,  actor_classic_tool_method_name
              end

              define_method(actor_raw_tool_method_name) do |options={}|
                Patriarch::ToolServices::RedisExtractorService.instance.get_ids_from_sorted_set(self,redis_key,options.merge({:tripartite => true , :protagonist_type  => :actor}))
              end
              if options[:as]
                progressive_present_relation_type_alias = (Verbs::Conjugator.conjugate options[:as], :aspect => :progressive).split(/ /).last
                alias_for_actor_raw = actor_raw_tool_method_name.sub(progressive_present_relation_type,progressive_present_relation_type_alias)
                alias_method alias_for_actor_raw,  actor_raw_tool_method_name
              end

              define_method(medium_classic_tool_method_name) do |options={}|
                Patriarch::ToolServices::RedisExtractorService.instance.
                  get_models_from_ids(self,via_model.to_s.classify.constantize,medium_raw_tool_method_name,options.merge({:tripartite => true}))
              end
              #TODO add alias there
              # FIXME id in it is wrong ...
              define_method(medium_raw_tool_method_name) do |options={}|
                Patriarch::ToolServices::RedisExtractorService.instance.get_ids_from_sorted_set(self,redis_key,options.merge({:tripartite => true , :protagonist_type  => :medium}))
              end

            end
          end

        end
      end

      #:nodoc:
      def included(klass)
        # Classic extend of ClassMethods module embedded within this module
        klass.extend ClassMethods
        unless klass.respond_to?(:before_destroy) && klass.respond_to?(:after_destroy)
          raise AddBehaviourDependencyError, "class #{klass.name} including Patriarch::Behaviours does not support callbacks"
        end

        # insert callback logic here : behaviours have to be cleaned when a destroy occur in SQL
        # First we build the transaction with #compute_redis_dependencies
        # We execute the clean only when the SQL transaction commit to avoid data inconsistency if a SQL Rollback occurs
        # in a middle of a cascading transaction
        klass.class_eval do

          before_destroy :compute_redis_dependencies
          def compute_redis_dependencies
            @redis_destroy_transac_object = Patriarch::ToolServices::RedisCleanerService.instance.clean_all(self)
          end

          after_commit  :clean_redis_dependencies
          def clean_redis_dependencies
            unless @redis_destroy_transac_object.nil?
              @redis_destroy_transac_object.execute
            end
            @redis_destroy_transac_object = nil
          end

          after_rollback :clean_redis_transac_object
          def clean_redis_transac_object
            @redis_destroy_transac_object = nil
          end

        end

        class << klass;
          # not sure if should add an alias like behaviours here
          # might override other behaviour methods from other namespaces
          # but patriarch is damn heavy here
          attr_accessor :patriarch_behaviours
        end
      end
    end

    # Gathers class methods that should be available for model that include Patriarch::Behaviours
    module ClassMethods


      # Helper that checks if options syntax is correct
      # Raises errors if not with indications on what should be corrected
      # @param [Object] options
      def check_add_behaviour_syntax(options)
        if options[:medium_between] || options[:via]
          check_tripartite_add_behaviour_syntax(options)
        elsif options[:on] || options[:by]
          check_bipartite_add_behaviour_syntax(options)
        end
      end

      #:nodoc:
      def check_bipartite_add_behaviour_syntax(options)
        # Either you add a behaviour on another model or allow a behaviour on you by another model
        unless options[:by].nil? ^ options[:on].nil?
          raise AddBehaviourSyntaxError, "you must not define a behaviour as active (using on) and passive at the same time (using by)"
        end

        # as option should be a string or symbol
        if options[:as]
          unless [Symbol,String].include?(options[:as].class)
            raise AddBehaviourSyntaxError, "as option should be a string or symbol"
          end
        end
      end

      #:nodoc:
      def check_tripartite_add_behaviour_syntax(options)
        # Xor on options :medium_between and :via, disparate cases ...
        unless options[:medium_between].nil? ^ options[:via].nil?
          raise AddBehaviourSyntaxError, "you must not define a behaviour as active (using on) and passive at the same time (using by)"
        end

        if options[:via]
          unless options[:by].nil? ^ options[:on].nil?
            raise AddBehaviourSyntaxError, "you must not define a behaviour as active (using on) and passive at the same time (using by)"
          end
        else
          medium_between_option = options[:medium_between]
          unless medium_between_option.is_a?(Array) && medium_between_option.size == 2
            raise AddBehaviourSyntaxError, "syntax with medium medium_between requires that you provide an array of two symbols/strings"
          end
        end

        if options[:as]
          # as option should be a string or symbol
          unless [Symbol,String].include?(options[:as].class)
            raise AddBehaviourSyntaxError, "as option should be a string or symbol"
          end
        end
      end

      def add_behaviour(behaviour,options)
        if options[:medium_between] || options[:via]
          add_tripartite_behaviour(behaviour,options)
        elsif options[:on] || options[:by]
          add_bipartite_behaviour(behaviour,options)
        else
          raise AddBehaviourSyntaxError, "syntax is wrong, declaration does not suplly enough arguments declared with right options"
        end
      end

      # @param [Module] module_to_complete
      # @param [String] behaviour
      # @param [Object] options
      # Adds alias for behaviour calls available to models if options allow to do so
      def add_aliases_for_functionality(module_to_complete,behaviour,options)
        if options[:as]
          behaviour_alias = options[:as].to_s
        end

        if options[:undo_as]
          behaviour_undo_alias = options[:undo_as].to_s
        end

        if options[:on]
          module_to_complete.class_eval do
            # add behaviour_alias
            if options[:as]
              alias_method behaviour_alias, behaviour
            end

            # add undo_behaviour_alias
            if options[:undo_as]
              alias_method behaviour_undo_alias, Patriarch.undo(behaviour)
            end
          end
        end
      end

      # @param [Module] anon_module
      # @param [String] behaviour
      # Add basic behaviours calls to models in bipartite situations
      def add_functionality_for_bipartite(anon_module,behaviour)
        anon_module.class_eval do

          # use define_method to get otherwise out of scope variables
          # usage of self.send(:define_method,:included) would fail as it would declare included
          # as a method available to class including this module not as an override of the included
          # hook
          (class << self; self; end).send(:define_method,:included) do |klass|
            klass.extend ActiveModel::Callbacks
            klass.send(:define_model_callbacks, behaviour, Patriarch.undo(behaviour))
          end

          # behave
          define_method(behaviour) do |entity,options={}|
            run_callbacks behaviour do
              "Patriarch::Services::#{behaviour.classify}::ManagerService".constantize.instance.resolve(self,entity,options)
            end
          end

          # undo_behave
          define_method(Patriarch.undo(behaviour).to_sym) do |entity,options={}|
            run_callbacks Patriarch.undo(behaviour).to_sym do
              "Patriarch::Services::#{(Patriarch.undo(behaviour)).classify}::ManagerService".constantize.instance.resolve(self,entity,options)
            end
          end

        end # class_eval
      end

      # @param [Module] anon_module
      # @param [String] behaviour
      # Add basic behaviours calls to models in tripartite situations
      def add_functionality_for_tripartite(anon_module,behaviour)
        anon_module.class_eval do

          # use define_method to get otherwise out of scope variables
          # usage of self.send(:define_method,:included) would fail as it would declare included
          # as a method available to class including this module not as an override of the included
          # hook
          (class << self; self; end).send(:define_method,:included) do |klass|
            klass.extend ActiveModel::Callbacks
            klass.send(:define_model_callbacks, behaviour, Patriarch.undo(behaviour))
          end

          # behave
          define_method(behaviour) do |via_entity,entity,options={}|
            run_callbacks behaviour.to_sym do
              "Patriarch::Services::#{behaviour.classify}::ManagerService".constantize.instance.resolve(self,entity,via_entity,options)
            end
          end

          # undo_behave
          define_method(Patriarch.undo(behaviour)) do |via_entity,entity,options={}|
            run_callbacks Patriarch.undo(behaviour).to_sym do
              "Patriarch::Services::#{Patriarch.undo(behaviour).classify}::ManagerService".constantize.instance.resolve(self,entity,via_entity,options)
            end
          end

        end # class_eval
      end

      # @param [String] behaviour
      # @param [Hash] options
      # Registers at a class level a declaration of bipartite behaviour so that it can be used
      # to check protagonist types when calling behaviours
      def register_bipartite_behaviour(behaviour,options)
        behaviour = behaviour.underscore.to_sym

        self.patriarch_behaviours ||= { }
        self.patriarch_behaviours[behaviour] ||= []

        Array(options[:on]).each do |target_class_sym|
          self.patriarch_behaviours[behaviour] << [self,target_class_sym.to_s.classify.constantize]
        end

        Array(options[:by]).each do |actor_class_sym|          
          self.patriarch_behaviours[behaviour] << [actor_class_sym.to_s.classify.constantize,self]
        end
      end

      # @param [String] behaviour
      # @param [Hash] options
      # Registers at a class level a declaration of tripartite behaviour so that it can be used
      # to check protagonist types when calling behaviours
      def register_tripartite_behaviour(behaviour,options)
        behaviour = behaviour.underscore.to_sym

        self.patriarch_behaviours ||= { }
        self.patriarch_behaviours[behaviour]  ||= []

        Array(options[:via]).each do |medium_class_sym|
          Array(options[:on]).each do |target_class_sym|
            target_class = target_class_sym.to_s.classify.constantize
            medium_class = medium_class_sym.to_s.classify.constantize
            self.patriarch_behaviours[behaviour] << [self, target_class, medium_class]
          end

          Array(options[:by]).each do |actor_class_sym|
            actor_class  = actor_class_sym.to_s.classify.constantize
            medium_class = medium_class_sym.to_s.classify.constantize
            self.patriarch_behaviours[behaviour] << [actor_class, self, medium_class]
          end
        end

        # FIXME should support enumerations of medium betweens like this
        # [[actor1,target1],[actor2,target2]]
        if options[:medium_between]
          actor_class  = options[:medium_between].first.to_s.classify.constantize
          target_class = options[:medium_between].last.to_s.classify.constantize
          self.patriarch_behaviours[behaviour] << [actor_class, target_class, self]
        end
      end

      # @param [Symbol] behaviour the tripartite behaviour we want to add to our model
      # @param [Object] options information useful to determine if supplementary functionality such as aliases or custom
      # tool method shall be implemented
      def add_tripartite_behaviour(behaviour,options)
        behaviour = behaviour.to_s.underscore

        check_add_behaviour_syntax(options)

        methods_mod = Module.new do; end

        # Target on Actor cases
        if options[:via]
          # Actor case
          if options[:on]
            add_functionality_for_tripartite(methods_mod,behaviour)
            add_aliases_for_functionality(methods_mod,behaviour,options)
            acted_on_model_list = Array(options[:on])
            via_model_list = Array(options[:via])
            Patriarch::Behaviours.complete_custom_active_module_tripartite(methods_mod,behaviour,acted_on_model_list,via_model_list,options)
            #Target case
          elsif options[:by]
            targeted_by_model_list = Array(options[:by])
            via_model_list = Array(options[:via])
            Patriarch::Behaviours.complete_custom_passive_module_tripartite(methods_mod,behaviour,targeted_by_model_list,via_model_list,options)
          end
        # Medium case
        elsif options[:medium_between]
          # Define tool_methods_here ...
        end

        # Finally ...
        include methods_mod
        # include in which we can override the methods of the previous custom module included there ...
        include "Patriarch::Behaviours::#{behaviour.classify}::ToolsMethods".constantize
        # register behaviour we just added
        register_tripartite_behaviour(behaviour,options)
      end

      # @param [Symbol] behaviour the bipartite behaviour we want to add to our model
      # @param [Object] options information useful to determine if supplementary functionality such as aliases or custom
      # tool method shall be implemented
      def add_bipartite_behaviour(behaviour,options)
        # add_behaviour :like, :by => bla,  :on => [:item,:user] || :community, :as => :love, :reverse_as => :hate ...

        behaviour = behaviour.to_s.underscore

        check_add_behaviour_syntax(options)

        methods_mod = Module.new do; end

        if options[:on]
          # Adds active methods and defines the hook to set callbacks on them
          add_functionality_for_bipartite(methods_mod,behaviour)
          add_aliases_for_functionality(methods_mod,behaviour,options)
          acted_on_model_list = Array(options[:on])
          Patriarch::Behaviours.complete_custom_active_module_bipartite(methods_mod,behaviour,acted_on_model_list,options)
        else
          targeted_by_model_list = Array(options[:by])
          Patriarch::Behaviours.complete_custom_passive_module_bipartite(methods_mod,behaviour,targeted_by_model_list,options)
        end

        # Finally includes the custom module
        include methods_mod
        # include in which we can override the methods of the previous custom module included there ...
        include "Patriarch::Behaviours::#{behaviour.classify}::ToolsMethods".constantize
        # register the behaviour we just added
        register_bipartite_behaviour(behaviour,options)
      end
    end # module Class Methods

  end # Behaviours
end # Patriarch
