require 'redis/objects/sorted_sets'

class InvalidRedisTypeException < Exception
end

# Services managing transactions must not know how the database layer is structured.
# We thus provide DAO through this RetrieverService and let them know about the interface of the DAOs
class Patriarch::DAOServices::RetrieverService < Patriarch::Service

  # @return [Hash] hash containing DAOs for RelationshipBuilderServices to toy with
  # @param  [Patriarch::Transaction] transaction_item
  # Only public method of this service, does the work of retrieving DAOs for a given transaction step
  def call(transaction_item)
    result = {}
    result[:actor]  = instantiate_DAO_for_actor(transaction_item)
    result[:target] = instantiate_DAO_for_target(transaction_item)
    if transaction_item.tripartite?
      result[:medium] = instantiate_DAO_for_medium(transaction_item)
    end
    result
  end

  protected
  #:nodoc:
  def get_DAO_info_for_actor(transaction_item)
    Patriarch::DAOServices::RedisMapperService.instance.call(transaction_item,:actor)
  end

  #:nodoc
    def get_DAO_info_for_target(transaction_item)
    Patriarch::DAOServices::RedisMapperService.instance.call(transaction_item,:target)
  end

  #:nodoc
  def get_DAO_info_for_medium(transaction_item)
    Patriarch::DAOServices::RedisMapperService.instance.call(transaction_item,:medium)
  end

  #:nodoc:
  def instantiate_DAO_for_actor(transaction_item)
    dao_info = get_DAO_info_for_actor(transaction_item)
    actor_id = transaction_item.actor_id
    actor_type = transaction_item.actor_type

    if Redis::Objects.constants.map(&:to_s).include?("#{dao_info[:type].pluralize.camelize}")
      redis_dao_class = "Redis::#{dao_info[:type].camelize}".constantize
    else
      raise InvalidRedisTypeException, "Redis constants available were #{Redis::Objects.constants} and supplied type was #{dao_info[:type]}"
    end

    redis_dao_class.new("#{actor_type}:#{actor_id}:#{dao_info[:key]}",dao_info[:options])
  end


  #:nodoc:
  def instantiate_DAO_for_medium(transaction_item)
    dao_info = get_DAO_info_for_medium(transaction_item)
    medium_id = transaction_item.medium_id
    medium_type = transaction_item.medium_type

    if Redis::Objects.constants.map(&:to_s).include?("#{dao_info[:type].pluralize.camelize}")
      redis_dao_class = "Redis::#{dao_info[:type].camelize}".constantize
    else
      raise InvalidRedisTypeException, "Redis constants available were #{Redis::Objects.constants} and supplied type was #{dao_info[:type]}"
    end

    redis_dao_class.new("#{medium_type}:#{medium_id}:#{dao_info[:key]}",dao_info[:options])
  end


  #:nodoc:
  def instantiate_DAO_for_target(transaction_item)
    dao_info = get_DAO_info_for_target(transaction_item)
    target_id = transaction_item.target_id
    target_type = transaction_item.target_type

    if Redis::Objects.constants.map(&:to_s).include?("#{dao_info[:type].pluralize.camelize}")
      redis_dao_class = "Redis::#{dao_info[:type].camelize}".constantize
    else
      raise InvalidRedisTypeException, "Redis constants available were #{Redis::Objects.constants} and supplied type was #{dao_info[:type]}"
    end
    redis_dao_class.new("#{target_type}:#{target_id}:#{dao_info[:key]}",dao_info[:options])
  end
end

