# Adding behaviours to classes through Patriarch induces writes in Redis Database
# What happens when we want to destroy an item is that it fires ActiveRecord::Callbacks to deal
# with this kind of stuff. We build a transaction item whose execute would clean redis stuff
# but do nothing until commit happens in case SQL aborts.
# This service takes care of building clean items
class Patriarch::ToolServices::RedisCleanerService < Patriarch::Service

  # @param [Object] calling_entity the entity calling the clean upon itself
  # @param [Object] options various options we may want to pass, used here to transmit a transaction item if a clean transaction wraps all the clean
  # @return [Patriarch::Transaction] returns a transaction item that is either built from scratch or was passed through options and has been completed
  def clean_all(calling_entity,options={})
    transac = options[:transac] || Patriarch::TransactionServices::TransactionManagerService.instance.new_transaction("clean_all".to_sym)
    options = { :transac => transac }

    calling_entity.class.patriarch_behaviours.keys.each do |behaviour|
      transac.steps << clean_behaviour(calling_entity,behaviour)
    end

    transac
  end

  # @return [Patriarch::Transaction] returns a transaction item that is either built from scratch or was passed through options and has been completed
  # @param [Object] calling_entity the entity calling the clean upon itself
  # @param [Object] behaviour the behaviour the entity wants to clean
  # @param [Object] options various options we may want to pass, used here to transmit a transaction item if a clean transaction wraps all the clean
  # behaviours so it runs once and not multiple transactions.
  def clean_behaviour(calling_entity,behaviour,options={})
    my_behaviours = calling_entity.class.patriarch_behaviours
    behaviour_symbol = behaviour.to_s.underscore.to_sym

    if my_behaviours.include? behaviour_symbol
      if my_behaviours[behaviour_symbol].first.size == 3
        clean_tripartite_behaviour(calling_entity,behaviour,options)
      elsif my_behaviours[behaviour_symbol].first.size == 2
        clean_bipartite_behaviour(calling_entity,behaviour,options)
      else
        raise "Bad Behaviour declaration occured, no doublet or triplet are registered"
      end
    end
  end


  #:nodoc:
  def clean_tripartite_behaviour(calling_entity,behaviour,options={})
    my_class = calling_entity.class
    tripartite_behaviour = my_class.patriarch_behaviours[behaviour.to_s.underscore.to_sym]

    transac = options[:transac] || Patriarch::TransactionServices::TransactionManagerService.instance.new_transaction("clean_#{behaviour.to_s}".to_sym)
    options = { :transac => transac, :stop_execution => true}
      
    tripartite_behaviour.each do |protagonists|
      clean_tripartite_behaviour_for_actor(calling_entity,behaviour,protagonists,options)  if protagonists.first  == my_class
      clean_tripartite_behaviour_for_target(calling_entity,behaviour,protagonists,options) if protagonists.second == my_class
      clean_tripartite_behaviour_for_medium(calling_entity,behaviour,protagonists,options) if protagonists.third  == my_class
    end      

    transac
  end #clean_tripartite_behaviour

  #:nodoc:
  def clean_bipartite_behaviour(calling_entity,behaviour,options={})
    my_class = calling_entity.class
    bipartite_behaviour = my_class.patriarch_behaviours[behaviour.to_s.underscore.to_sym]

    transac = options[:transac] || Patriarch::TransactionServices::TransactionManagerService.instance.new_transaction("clean_#{behaviour.to_s}".to_sym)
    options = { :transac => transac, :stop_execution => true}

    bipartite_behaviour.each do |protagonists|
      clean_bipartite_behaviour_for_actor(calling_entity,behaviour,protagonists,options)  if protagonists.first  == my_class
      clean_bipartite_behaviour_for_target(calling_entity,behaviour,protagonists,options) if protagonists.second == my_class
    end  

    transac
  end # #clean_bipartite_behaviour

  protected

  def clean_bipartite_behaviour_for_actor(calling_entity,behaviour,protagonists,options)
    acted_on_model = protagonists.second
    calling_entity.send("#{acted_on_model.to_s.tableize}_i_#{behaviour.to_s}").each do |target_i_behaved_on|
      calling_entity.send(Patriarch.undo(behaviour.to_s).underscore,target_i_behaved_on,options)
    end   
  end

  def clean_bipartite_behaviour_for_target(calling_entity,behaviour,protagonists,options)
    # Dealing with passive behaviour
    behaviour_symbol = behaviour.to_s.underscore.to_sym

    # Compute names based on conventions
    progressive_present_behaviour =
      (Verbs::Conjugator.conjugate behaviour_symbol, :aspect => :progressive).split(/ /).last

    targetted_by_model = protagonists.first
    calling_entity.send("#{targetted_by_model.to_s.tableize}_#{progressive_present_behaviour}_me").each do |actor_that_behaved_on_me|
      actor_that_behaved_on_me.send(Patriarch.undo(behaviour.to_s).underscore,calling_entity,options)
    end    
  end

  def clean_tripartite_behaviour_for_actor(calling_entity,behaviour,protagonists,options)
    # Dealing with active behaviour
    acted_on_model = protagonists.second
    via_model      = protagonists.third
    calling_entity.send("#{acted_on_model.to_s.tableize}_i_#{behaviour.to_s}_via_#{via_model.to_s.tableize}", :with_medium => true).each do |tripartite_hash|
      target_i_behaved_on  = Array(acted_on_model.find(tripartite_hash[:target])).first
      medium_for_behaviour = Array(via_model.find(tripartite_hash[:medium])).first
      calling_entity.send(Patriarch.undo(behaviour.to_s).underscore,medium_for_behaviour,target_i_behaved_on,options)
    end
  end

  def clean_tripartite_behaviour_for_target(calling_entity,behaviour,protagonists,options)
    # Dealing with passive behaviour
    behaviour_symbol = behaviour.to_s.underscore.to_sym

    # Compute names based on conventions
    progressive_present_behaviour =
      (Verbs::Conjugator.conjugate behaviour_symbol, :aspect => :progressive).split(/ /).last

    targetted_by_model = protagonists.first
    via_model          = protagonists.third
    calling_entity.send("#{targetted_by_model.to_s.tableize}_#{progressive_present_behaviour}_me_via_#{via_model.to_s.tableize}", :with_medium => true).each do |tripartite_hash|
      actor_that_behaved_on_me  = Array(targetted_by_model.find(tripartite_hash[:actor])).first
      medium_for_behaviour = Array(via_model.find(tripartite_hash[:medium])).first
      actor_that_behaved_on_me.send(Patriarch.undo(behaviour.to_s).underscore,medium_for_behaviour,calling_entity,options)
    end
  end 

  def clean_tripartite_behaviour_for_medium(calling_entity,behaviour,protagonists,options)
    # do nothing. Not implemented yet.
    # TODO sort this bizarre thing
    # Previous version comment said => #do nothing, we do not take care of data relative to medium ...
  end

end  # class Patriarch ...