module Patriarch 
  # Represents a "step" i.e. what should occur from no sql database point of view when a model instance 
  # invokes a behaviour. It is encapsulated into Patriarch::Transaction item and is never exposed.
  # When needed, Patriarch::Transaction will forward some method call to the current TransactionStep
  # that is being built
  class TransactionStep
    attr_accessor :context, :queue

    def initialize(relation_type,actor,target,medium = nil)
      @context = {
        :relation_type  => relation_type,
        :actor_type     => actor.class.name.underscore.to_sym,
        :target_type    => target.class.name.underscore.to_sym,
        :actor_id       => actor.id,
        :target_id      => target.id,
      }
      if medium
        @context.merge! :medium_id => medium.id, :medium_type => medium.class.name.underscore.to_sym
      end
      @queue = []
    end

    # defines access methods to context fields letting it be encapsulated nicely
    [:relation_type,:actor_type,:target_type,:medium_type,:actor_id,:target_id,:medium_id].each do |context_key|
      define_method(context_key) do
        context[context_key]
      end
    end

    # @return the medium of this step if it exists
    def medium
      return nil unless medium_id && medium_type
      medium_type.to_s.camelize.constantize.find medium_id
    end

    # @return the actor of this step
    def actor
      actor_type.to_s.camelize.constantize.find actor_id
    end

    # @return the target of this step
    def target
      target_type.to_s.camelize.constantize.find target_id
    end

    def protagonists_models 
      # By convention always actor, target, model for tripartite and 
      # actor, model for bipartite
      [actor,target,medium].compact.map(&:class)
    end

    # execute the redis instructions embedded into this step against
    # instructions are a dao, a method to call upon and its argument embedded into a OpenStruct
    # before Patriarch 0.2.5, was embedded directly in a proc, was a complete failure for unknown reasons
    # Patriach::Transaction use this method on each of the steps it registered within a multi block
    def execute
      queue.each do |redis_struct_instruction|
        redis_struct_instruction.execute
      end
    end

    # @param [Patriarch::RedisInstruction] ostruct dao with parameters and method to call embedded here to be processed later
    def add_to_queue(struct)
      queue << struct
    end
  end
end
